"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = void 0;
function _toConsumableArray(r) { return _arrayWithoutHoles(r) || _iterableToArray(r) || _unsupportedIterableToArray(r) || _nonIterableSpread(); }
function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); }
function _iterableToArray(r) { if ("undefined" != typeof Symbol && null != r[Symbol.iterator] || null != r["@@iterator"]) return Array.from(r); }
function _arrayWithoutHoles(r) { if (Array.isArray(r)) return _arrayLikeToArray(r); }
function _createForOfIteratorHelper(r, e) { var t = "undefined" != typeof Symbol && r[Symbol.iterator] || r["@@iterator"]; if (!t) { if (Array.isArray(r) || (t = _unsupportedIterableToArray(r)) || e && r && "number" == typeof r.length) { t && (r = t); var _n = 0, F = function F() {}; return { s: F, n: function n() { return _n >= r.length ? { done: !0 } : { done: !1, value: r[_n++] }; }, e: function e(r) { throw r; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var o, a = !0, u = !1; return { s: function s() { t = t.call(r); }, n: function n() { var r = t.next(); return a = r.done, r; }, e: function e(r) { u = !0, o = r; }, f: function f() { try { a || null == t["return"] || t["return"](); } finally { if (u) throw o; } } }; }
function _unsupportedIterableToArray(r, a) { if (r) { if ("string" == typeof r) return _arrayLikeToArray(r, a); var t = {}.toString.call(r).slice(8, -1); return "Object" === t && r.constructor && (t = r.constructor.name), "Map" === t || "Set" === t ? Array.from(r) : "Arguments" === t || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(t) ? _arrayLikeToArray(r, a) : void 0; } }
function _arrayLikeToArray(r, a) { (null == a || a > r.length) && (a = r.length); for (var e = 0, n = Array(a); e < a; e++) n[e] = r[e]; return n; }
/**
 * @typedef Options
 * @prop {string} functionName name of imported function
 * @prop {string} moduleName name of module to import function from
 */

/**
 * Checks for function name at import declaration.
 * @param {import('@babel/types').ImportDeclaration} importDeclaration
 * @param {Options} options
 * @returns {boolean}
 */
var getHasFunctionNameInImport = function getHasFunctionNameInImport(importDeclaration, _ref) {
  var functionName = _ref.functionName;
  var _iterator = _createForOfIteratorHelper(importDeclaration.specifiers),
    _step;
  try {
    for (_iterator.s(); !(_step = _iterator.n()).done;) {
      var specifier = _step.value;
      if (specifier.type === 'ImportSpecifier') {
        if (specifier.imported.type === 'StringLiteral') {
          if (specifier.imported.value === functionName) {
            return true;
          }
        } else if (specifier.imported.type === 'Identifier' && specifier.imported.name === functionName) {
          return true;
        }
      }
    }
  } catch (err) {
    _iterator.e(err);
  } finally {
    _iterator.f();
  }
  return false;
};

/**
 * Returns index of statement with function name from options or -1 otherwise.
 * @param {import('@babel/types').Statement[]} statements
 * @param {Options} options
 * @returns {number}
 */
var findIndexOfStatement = function findIndexOfStatement(statements, _ref2) {
  var functionName = _ref2.functionName;
  for (var index = 0; index < statements.length; index += 1) {
    var statement = statements[index];
    if (statement.type === 'ExpressionStatement') {
      var expression = statement.expression;
      if (expression.type === 'CallExpression') {
        var callee = expression.callee;
        if (callee.type === 'Identifier') {
          if (callee.name === functionName) {
            return index;
          }
        }
      }
    }
  }
  return -1;
};

/**
 * Wraps statements in try/catch and wraps results in function calls.
 * @param {import('@babel/core').NodePath<import('@babel/types').Function>} funcPath
 * @param {Options} options
 * @param {import('@babel/types')} t
 * @returns {void}
 */
var handleFunctionPath = function handleFunctionPath(funcPath, options, t) {
  var blockStatement = funcPath.node.body;
  if (blockStatement.type !== 'BlockStatement') {
    return;
  }
  var statementIndex = findIndexOfStatement(blockStatement.body, options);
  if (statementIndex === -1) {
    return;
  }
  var statement = blockStatement.body[statementIndex];
  if (statement.type !== 'ExpressionStatement') {
    return;
  }
  var endId = funcPath.scope.generateUidIdentifier('end');
  blockStatement.body[statementIndex] = t.variableDeclaration('const', [t.variableDeclarator(endId, statement.expression)]);
  var otherBlockStatement = t.blockStatement(blockStatement.body.slice(statementIndex + 1));
  var errorId = funcPath.scope.generateUidIdentifier('e');
  blockStatement.body = [].concat(_toConsumableArray(blockStatement.body.slice(0, statementIndex + 1)), [t.tryStatement(otherBlockStatement, t.catchClause(errorId, t.blockStatement([t.throwStatement(t.callExpression(endId, [errorId]))])))]);
  funcPath.traverse({
    ReturnStatement: function ReturnStatement(returnPath) {
      if (returnPath.scope.getFunctionParent() === funcPath.scope) {
        var isChildOfOtherBlockStatement = returnPath.findParent(function (path) {
          return path.node === otherBlockStatement;
        });
        if (isChildOfOtherBlockStatement) {
          var node = returnPath.node;
          node.argument = t.callExpression(endId, node.argument ? [node.argument] : []);
        }
      }
    }
  });
  var lastStatement = otherBlockStatement.body[otherBlockStatement.body.length - 1];
  if (lastStatement && lastStatement.type !== 'ReturnStatement') {
    otherBlockStatement.body.push(t.expressionStatement(t.callExpression(endId, [])));
  } else if (!lastStatement) {
    otherBlockStatement.body.push(t.expressionStatement(t.callExpression(endId, [])));
  }
};

/**
 * Wraps statements after first call of specified imported function in try/catch and then wraps each
 * returned result inside that try/catch block in result of that call.
 * @param {import('@babel/core')} api
 * @param {Options} options
 * @returns {import('@babel/core').PluginObj}
 */
var plugin = function plugin(_ref3, options) {
  var t = _ref3.types;
  return {
    name: 'babel-plugin-function-tracker',
    visitor: {
      Program: function Program(programPath) {
        var functionImport = programPath.node.body.find(function (node) {
          return node.type === 'ImportDeclaration' && node.source.value === options.moduleName && getHasFunctionNameInImport(node, options);
        });
        if (functionImport) {
          programPath.traverse({
            Function: function Function(functionPath) {
              handleFunctionPath(functionPath, options, t);
            }
          });
        }
      }
    }
  };
};
var _default = exports["default"] = plugin;